from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.binarytreeaction import BinaryTreeAction
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.data.collection.trees.comparison.comparator import Comparison, Equal, Left, Replace, Right
from rainbowadn.hashing.hashpoint import HashPoint

__all__ = ('AddAction', 'RemoveAction', 'ContainsAction',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class AddAction(
    BinaryTreeAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, key: HashPoint[ActiveKeyType]):
        assert isinstance(key, HashPoint)
        self.key = key

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return protocol.tree(tree, tree, self.key)

    def on_split(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            treel: TreeType,
            key: HashPoint[ActiveKeyType],
            treer: TreeType
    ) -> TreeType:
        comparison: Comparison = protocol.comparator.compare(key, self.key)
        assert isinstance(comparison, Comparison)
        if isinstance(comparison, Replace):
            return protocol.tree(treel, treer, self.key)
        elif isinstance(comparison, Left):
            return protocol.tree(self.on(protocol, treel), treer, key)
        elif isinstance(comparison, Right):
            return protocol.tree(treel, self.on(protocol, treer), key)
        else:
            raise TypeError


class MergeAction(
    BinaryTreeAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, treer: TreeType):
        self.treer = treer

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return self.treer

    def on_split(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            treel: TreeType,
            key: HashPoint[ActiveKeyType],
            treer: TreeType
    ) -> TreeType:
        return protocol.tree(
            MergeAction(treer).on(protocol, treel),
            self.treer,
            key
        )


class RemoveAction(
    BinaryTreeAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, key: HashPoint[ActiveKeyType]):
        assert isinstance(key, HashPoint)
        self.key = key

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return tree

    def on_split(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            treel: TreeType,
            key: HashPoint[ActiveKeyType],
            treer: TreeType
    ) -> TreeType:
        comparison: Comparison = protocol.comparator.compare(key, self.key)
        assert isinstance(comparison, Comparison)
        if isinstance(comparison, Equal):
            return MergeAction(treer).on(protocol, treel)
        elif isinstance(comparison, Left):
            return protocol.tree(self.on(protocol, treel), treer, key)
        elif isinstance(comparison, Right):
            return protocol.tree(treel, self.on(protocol, treer), key)
        else:
            raise TypeError


class ContainsAction(
    BinaryTreeAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        bool
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, key: HashPoint[ActiveKeyType]):
        assert isinstance(key, HashPoint)
        self.key = key

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> bool:
        return False

    def on_split(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            treel: TreeType,
            key: HashPoint[ActiveKeyType],
            treer: TreeType
    ) -> bool:
        comparison: Comparison = protocol.comparator.compare(key, self.key)
        assert isinstance(comparison, Comparison)
        if isinstance(comparison, Equal):
            return True
        elif isinstance(comparison, Left):
            return self.on(protocol, treel)
        elif isinstance(comparison, Right):
            return self.on(protocol, treer)
        else:
            raise TypeError
